/*
  Arduion TZXDuino/CASDuino tape amulator for Science of Cambridge MK14

  The circuit:
  - FLAG-0 attached to pin 2
  - pin 4 attached to SENSE-B and SIN
  - LED attached from pin 13 to ground (or use the built-in LED on most
    Arduino boards)

  created  11 Mar 2018
  modified 07 Apr 2018
  by Martin Lukasek

  This code is in the public domain.
*/

//load libraries
#include <Wire.h>
#include <LCD.h>
#include <LiquidCrystal_I2C.h>
#include <SD.h>

//Initialise the LCD
LiquidCrystal_I2C      lcd(0x27, 2, 1, 0, 4, 5, 6, 7);

// Variables
volatile int bitCounter = 0;   // counter for the number of bits
volatile int byteCounter = 0;   // counter for the number of bytes
volatile byte dataByte = 0;   // counter for the number of bytes
volatile int charPos = 0;   // counter for the postion on line

unsigned long timeDiff = 0;         //button debounce
char actFileName[32];
int actFileNr = 0;
int nrFiles = 0;
char dataBuf[256];
int dataSize = 0;

#define inputPin      2
#define outputPin     9             // TapeEmu pin 4
#define ledPin        13

#define btnPlay       17            //Play Button
#define btnStop       16            //Stop Button
#define btnUp         15            //Up button
#define btnDown       14            //Down button
#define btnRoot       7             //Return to SD card root

File root;

void getDirectory(File dir, int numTabs) {
  int i = 0;
  File entry =  dir.openNextFile();
  while( entry ) {
    if( !entry.isDirectory() ) {
      i++;
      // Serial.print(entry.name());
      // Serial.print("\t\t");
      // Serial.println(entry.size(), DEC);
    }
    entry.close();
    entry = dir.openNextFile();
  }
  entry.close();
  nrFiles=i;
  actFileNr=0;
  setFile(root, actFileNr);
}

void setFile(File dir, int fileNr) {
  int i = 0;
  dir.rewindDirectory();
  File entry = dir.openNextFile();
  while( entry && i<fileNr) {
    if( !entry.isDirectory() ) {
      i++;
    }
    entry.close();
    entry = dir.openNextFile();
  }
  if(i==fileNr)
    strncpy(actFileName, entry.name(), 32);
    // Serial.println(entry.name());
  entry.close();
}

void setup() {
  // initialize I/O pins
  pinMode(inputPin, INPUT);
  pinMode(outputPin, OUTPUT);
  // initialize the LED as an output
  pinMode(ledPin, OUTPUT);

  // initialize buttons with internal pullup 
  pinMode(btnPlay, INPUT_PULLUP);
  digitalWrite(btnPlay, HIGH);
  pinMode(btnStop, INPUT_PULLUP);
  digitalWrite(btnStop, HIGH);
  pinMode(btnUp, INPUT_PULLUP);
  digitalWrite(btnUp, HIGH);
  pinMode(btnDown, INPUT_PULLUP);
  digitalWrite(btnDown, HIGH);
  pinMode(btnRoot, INPUT_PULLUP);
  digitalWrite(btnRoot, HIGH); 

  // initialize serial communication:
  // Serial.begin(115200);
  // Serial.println(F("MK14Duino v0.3                 Martin Lukasek - www.8bity.cz"));
  // Serial.println();
  
  //Define the LCD as 16 column by 2 rows 
  lcd.begin (16,2);
  lcd.setBacklightPin(3, POSITIVE);
  lcd.setBacklight(HIGH);
  lcd.home(); // go home
  lcd.print(F("MK14Duino       "));

  // Serial.print(F("Initializing SD card..."));
  if (!SD.begin(10)) {
    // Serial.println(F("initialization failed! ABORTED"));
    lcd.setCursor(0, 1);
    lcd.print(F("SD Init FAILED!"));
    while (1);
  }
  // Serial.println(F("initialization done."));
  root = SD.open("/");
  getDirectory(root, 0);
  // Serial.println(F("Root folder read!"));
     
  // initalize interrupt routine
  // attachInterrupt(digitalPinToInterrupt(inputPin), readBit, RISING);
}

void readBit() {
  int i;
  for(i=0;i<100;i++)
    delayMicroseconds(100);
  dataByte = dataByte >> 1;
  if(digitalRead(inputPin) == HIGH) {
    dataByte = dataByte | B10000000;
  };
  bitCounter++;
  if(bitCounter == 8){
    if(dataByte < 16)
      // Serial.print("0");
    // Serial.print(dataByte,HEX);
    charPos+=2;
    if(charPos>31) {
      charPos=0;
      // Serial.println();
    }
    bitCounter=0;
    dataByte=0;
    byteCounter++;
  }
}

void sendByte(byte DTA) {
  int i,j;
  noInterrupts();
  for(i=0;i<8;i++) {
    for(j=0;j<4;j++) {
      digitalWrite(outputPin, HIGH);
      delayMicroseconds(500);
      digitalWrite(outputPin, LOW);
      delayMicroseconds(500);
    }
    if( ((DTA >> i) & B00000001) == 0) {
      digitalWrite(outputPin, LOW);
      for(j=0;j<24;j++)
        delayMicroseconds(500);
    } else {
      for(j=0;j<12;j++) {
        digitalWrite(outputPin, HIGH);
        delayMicroseconds(500);
        digitalWrite(outputPin, LOW);
        delayMicroseconds(500);
      }
    }
    digitalWrite(outputPin, LOW);
    for(j=0;j<160;j++)
      delayMicroseconds(100);
  }
  interrupts();
}

byte xtod(char c) {
  if (c>='0' && c<='9') return c-'0';
  if (c>='A' && c<='F') return c-'A'+10;
  if (c>='a' && c<='f') return c-'a'+10;
  return 0;        // not Hex digit
}

/*
void printHex(int num) {
  if(num<16)
    Serial.print("0");
  Serial.print(num, HEX);
} */

void playFile() {
  File df=SD.open(actFileName);
  char ch, chL, chH;
  byte rowLen;
  byte chkSum;
  byte chkSumRow;
  byte dataB;
  int i, rowNr;
  bool isHex=false;
  bool chkErr=false;
  if(df) {
    while( df.available() ) {
      // check whether binary or HEX data
      ch = df.read();
      chH = df.read();
      chL = df.read();
      df.seek(0);
      rowNr = 0;
      dataSize = 0;
      if(ch==':' && chH>='0' && chH<='2' && chL>='0' && chL<='F') {
        isHex=true;
        while( df.available() ) {
          ch = (char)df.read();
          if(ch==':') { // read line after :
            // Serial.print("ROW:");
            rowNr++;
            chkSum = 0;
            chH = (char)df.read();
            chL = (char)df.read();
            rowLen = 16*xtod(chH) + xtod(chL); 
            chkSum += rowLen;
            // Serial.print(">"); Serial.print(rowLen); Serial.print("<");
            chH = (char)df.read(); // ignore address
            chL = (char)df.read();
            chkSum += 16*xtod(chH) + xtod(chL);
            chH = (char)df.read();
            chL = (char)df.read();
            chkSum += 16*xtod(chH) + xtod(chL);
            chH = (char)df.read(); // ignore data type
            chL = (char)df.read();
            chkSum += 16*xtod(chH) + xtod(chL);
            // Serial.print("[");
            for(i=0; i<rowLen; i++) { // read all data bytes
              chH = (char)df.read();
              chL = (char)df.read();
              dataB = 16*xtod(chH) + xtod(chL);
              chkSum += dataB;
              // printHex(dataB);
              // sendByte(dataB);
              dataBuf[dataSize++]=dataB;
            }
            // Serial.print("]");
            chkSum = ~chkSum + 1;
            // Serial.print("*"); printHex(chkSum); Serial.print("*"); 
            chH = (char)df.read(); // read check sum
            chL = (char)df.read(); 
            chkSumRow = 16*xtod(chH) + xtod(chL); 
            // Serial.print("-"); printHex(chkSumRow); Serial.println("-");
            
            if(chkSumRow!=chkSum) {
              // Serial.print(F("Checksum ERROR on line "));
              // Serial.print(rowNr);
              // Serial.print(".");
              lcd.setCursor(0, 0);
              lcd.print(F("                "));
              lcd.setCursor(0, 0);
              lcd.print(F("ChkErr "));
              lcd.print(rowNr);
              lcd.print("-");
              lcd.print(chkSum,HEX);
              chkErr=true;
            }
          }
        }
      } else {
        isHex=false;
        // read binary file
        while (df.available()) {
          dataB = df.read();
          // sendByte(dataB);
          dataBuf[dataSize++]=dataB;
        }
      }

    }
    df.close();    

    if( !chkErr ) {
      lcd.setCursor(0, 0);
      lcd.print(F("Addr 007C GO -  "));
    }
    // wait 5s for 7c and GO on MK14
    for(i=5; i>0; i--) {
      lcd.setCursor(15, 0);
      lcd.print(i);
      delay(1000);
    }

    lcd.setCursor(0, 0);
    if( !chkErr ) {
      if( isHex )
        lcd.print(F("Playing HEX     "));
      else
        lcd.print(F("Playing BIN     "));
    }
    for (i=0; i<dataSize; i++) {
      sendByte(dataBuf[i]);
      lcd.setCursor(12, 0);
      lcd.print(100*i/dataSize);
      lcd.print("%  ");
    }
    // delay(1000);
    lcd.home();
    lcd.print(F("MK14Duino       "));
  } else {
    // read failed, not DF
  }
}
    
void doMenu() {
  int lastBtn, i;
  if( millis() - timeDiff > 50 ) {   // check switch every 50ms 
    timeDiff = millis();           // get current millisecond count
    if( digitalRead(btnPlay) == LOW ) {
      lastBtn=btnPlay;
      //Handle Play/Pause button
      // lcd.setCursor(13, 0);
      // lcd.print(F("Ply"));
      playFile();
    } else {
      if( digitalRead(btnStop) == LOW ) {
        lastBtn=btnStop;
        //Handle Stop button
        lcd.setCursor(13, 0);
        lcd.print(F("Stp"));
      } else {
        if( digitalRead(btnUp) == LOW ) {
          lastBtn=btnUp;
          //Handle Stop button
          if(actFileNr>0) {
            actFileNr--;
            setFile(root, actFileNr);
          }
          lcd.setCursor(13, 0);
          lcd.print(F("Up "));
        } else {
          if( digitalRead(btnDown) == LOW ) {
            lastBtn=btnDown;
            //Handle Stop button
            if(actFileNr<nrFiles-1) {
              actFileNr++;
              setFile(root, actFileNr);
            }
            lcd.setCursor(13, 0);
            lcd.print(F("Dwn"));
          } else {
            if( digitalRead(btnRoot) == LOW ) {
              lastBtn=btnRoot;
              //Handle Stop button
              lcd.setCursor(13, 0);
              lcd.print(F("Roo"));
            } else {
              lastBtn=0;
              lcd.setCursor(13, 0);
              lcd.print(F("0.3"));
            }
          }        
        }
      }
    }
    lcd.setCursor(0, 1);
    if(nrFiles && actFileNr<nrFiles) {
      lcd.print(actFileName);
      for(i=strlen(actFileName);i<16;i++)
        lcd.print(" ");
    } else {
      lcd.print(F("No files/SD card"));
    }
    if(lastBtn!=0) {
      while( digitalRead(lastBtn) == LOW ) {
        //prevent button repeats by waiting until the button is released.
        delay(50);
      }
    }
  }
}

void procSerInput() {
  int i;
  char c[2];
  byte dataVal;

  if( Serial.available()>1 ) {
    c[0] = Serial.read();
    c[1] = Serial.read();
    dataVal = 0;
    if( (c[0]>='0') && (c[0]<='F') && (c[1]>='0') && (c[1]<='F') ) {
      if(c[0]>='A')
        dataVal = (c[0]-'A'+10)*16;
      else
        dataVal = (c[0]-'0')*16;
      if(c[1]>='A')
        dataVal += c[1]-'A'+10;
      else
        dataVal += c[1]-'0';
      if(dataVal < 16)
        Serial.print("0");
      Serial.print(dataVal,HEX);
      charPos+=2;
      if(charPos>31) {
        charPos=0;
        Serial.println();
      }
      sendByte(dataVal);
    }
  }
}

void loop() {
    doMenu();
}
